import re

_PRIMITIVE_TYPES = {
    'B': 'byte',
    'C': 'char',
    'D': 'double',
    'F': 'float',
    'I': 'int',
    'J': 'long',
    'S': 'short',
    'Z': 'boolean'
}


class Package(object):
    def __init__(self, parts):
        self.parts = tuple(parts)
        self.name = '.'.join(self.parts)
        self.path = '/'.join(self.parts) + '/'

    def get_member_path(self, member):
        return self.path + member + '.class'

    def __eq__(self, other):
        return self.parts == other.parts

    def __ne__(self, other):
        return self.parts != other.parts

    def __hash__(self):
        return hash(self.parts)

    def __str__(self):
        return self.name


def parse_name(name, separator='.'):
    parts = name.split(separator)
    return (Package(parts[:-1]), parts[-1])


class LinkableClass(object):
    def __init__(self, class_info):
        self.package, self.name = parse_name(class_info.get_this(), '/')
        self.full_name = '{}.{}'.format(self.package, self.name)

        self.fields = tuple(LinkableField(f) for f in class_info.fields)

        methods = []
        for m in filter(is_linkable_method, class_info.methods):
            methods.append(LinkableMethod(self.name, m))
        self.methods = tuple(methods)

    def get_member(self, member):
        field = next((f for f in self.fields if member == f.name), None)
        if field:
            return field

        match = re.match(r'^(.+?)(?:\((.*)\))?$', member)
        if match:
            name, args = match.group(1, 2)
            method = next((m for m in self.methods if name == m.name), None)
            if method:
                if args is None:
                    return method

                if args.strip():
                    arglist = [a.strip() for a in args.split(',')]
                else:
                    arglist = []

                if method.has_args(arglist):
                    return method

        return None

    def __str__(self):
        return '{}.{}'.format(self.package.name, self.name)


class LinkableField(object):
    def __init__(self, field):
        self.name = field.get_name()

    def get_url_fragment(self):
        return self.name

    def __str__(self):
        return self.get_url_fragment()


class LinkableMethod(object):
    def __init__(self, class_name, method):
        name = method.get_name()
        if name == '<init>':
            self.name = class_name.split('$')[-1]
        else:
            self.name = name

        args = method.pretty_arg_types()
        arg_signatures = get_arg_signatures(method.get_signature())

        # 'None' extension behavior of map is important
        self.args = map(Argument, args, arg_signatures)
        if self.args:
            self.args[-1].vararg = method.is_varargs()

    def has_args(self, args):
        if len(args) != len(self.args):
            return False

        for arg, a in zip(args, self.args):
            if not a.endswith(arg):
                return False

        return True

    def get_url_fragment(self):
        args = ', '.join([str(a) for a in self.args])
        return '{}({})'.format(self.name, args)

    def __str__(self):
        return self.get_url_fragment()


class Argument(object):
    def __init__(self, arg, sig=None, vararg=False):
        self.parts = arg.split('.')
        if sig:
            self.sig_parts = sig.split('.')
        else:
            self.sig_parts = []
        self.vararg = vararg

    def endswith(self, arg):
        # standardize varargs to array syntax
        arg_parts = arg.replace('...', '[]').split('.')

        if len(arg_parts) > len(self.parts):
            return False

        endswith = self.parts[-len(arg_parts):] == arg_parts
        if not endswith:
            endswith = self.parts[-len(arg_parts):] == arg_parts

        if not endswith and self.sig_parts:
            endswith = self.sig_parts[-len(arg_parts):] == arg_parts

        return endswith

    def __str__(self):
        arg_str = '.'.join(self.sig_parts)
        if not arg_str:
            arg_str = '.'.join(self.parts)

        if self.vararg:
            # replace trailing '[]' with '...'
            return arg_str[:-2] + '...'
        else:
            return arg_str


def is_linkable_method(method_info):
    return not (method_info.is_bridge() or
                method_info.is_synthetic() or
                method_info.get_name() == '<clinit>')


def get_arg_signatures(sig):
    if sig is None:
        return []

    start = sig.find('(')
    if start < 0:
        return []

    args = []
    s = sig[start+1:sig.find(')', start)]
    while s:
        arg, s = _next_arg(s)
        args.append(arg)

    return args


# TODO there's an unimplemented stub in javatools to produce "pretty"
# signatures. When that is implemented, check if it can be used.
def _next_arg(s):
    c = s[0]
    if c in _PRIMITIVE_TYPES:
        return (_PRIMITIVE_TYPES[c], s[1:])
    elif c == 'T':
        i = s.find(';')
        return (s[1:i], s[i+1:])
    elif c == 'L':
        name, s = _next_arg_class(s[1:])
        return (name.replace('/', '.'), s)
    elif c == '[':
        t, s = _next_arg(s[1:])
        return ('{}[]'.format(t), s)
    else:
        raise ValueError('unknown arg type: {} in {}'.format(c, s))


def _next_arg_class(s):
    end = name_end = len(s)

    type_arg_count = 0
    for i in range(0, len(s)):
        c = s[i]
        if c == ';':
            name_end = min(i, name_end)
            if type_arg_count == 0:
                end = i + 1
                break
        elif c == '<':
            name_end = min(i, name_end)
            type_arg_count += 1
        elif c == '>':
            type_arg_count -= 1

    return (s[:name_end], s[end:])



import itertools
import javatools
import os
import zipfile

from javatools import ziputils

from .model import LinkableClass, Package, parse_name

def extract_class(jar, name):
    """Extracts a LinkableClass from a jar.

    Args:
        jar: An open ZipFile instance.
        name: A string containing the binary name of a class.

    Raises:
        KeyError: The class does not exist in the jar.
    """

    with jar.open(name) as entry:
        return LinkableClass(javatools.unpack_class(entry))


def is_jar(path):
    return path.endswith('.jar') and zipfile.is_zipfile(path)


def expand_classpath_entry(entry):
    if os.path.isdir(entry) or is_jar(entry):
        return [entry]
    elif os.path.basename(entry) == '*':
        entry = os.path.dirname(entry)
        contents = [os.path.join(entry, c) for c in os.listdir(entry)]
        return [c for c in contents if is_jar(c)]
    else:
        raise ValueError('Invalid classpath entry: {}'.format(entry))


def open_classpath_entry(entry):
    if os.path.isdir(entry):
        return ExplodedZipFile(entry)
    elif is_jar(entry):
        return zipfile.ZipFile(entry, 'r')
    else:
        raise ValueError('Invalid classpath entry: {}'.format(entry))


class ClassLoader(object):
    def __init__(self, paths):
        entries = [expand_classpath_entry(p) for p in paths]
        self.entries = list(itertools.chain.from_iterable(entries)) # flatten

        # {Package : {class name : LinkableClass}}
        self.packages = {}

    def load(self, name):
        package, class_name = parse_name(name)

        try:
            return self.packages[package][class_name]
        except KeyError:
            clazz = self.find(name)
            if clazz:
                if clazz.package != package or clazz.name != class_name:
                    msg = "Wanted class '{}', but '{}' was loaded"
                    raise ValueError(msg.format(name, clazz))

                classes = self.packages.setdefault(package, {})
                classes[class_name] = clazz

            return clazz

    def find(self, name):
        package, class_name = parse_name(name)
        path = package.get_member_path(class_name)

        for entry in self.entries:
            with open_classpath_entry(entry) as jar:
                try:
                    return extract_class(jar, path)
                except KeyError:
                    pass

        return None

    # TODO take either a Package or a string name
    def find_package(self, name):
        package = Package(name.split('.'))

        if package in self.packages:
            return package

        for entry in self.entries:
            with open_classpath_entry(entry) as jar:
                try:
                    jar.getinfo(package.path)
                except KeyError:
                    pass
                else:
                    # TODO avoid changing state in find method
                    self.packages[package] = {}
                    return package

        return None


class ExplodedZipFile(ziputils.ExplodedZipFile):
    """A ZipFile-like object that wraps a directory.

    Changes the behavior of javatools.ziputils.ExplodedZipFile to be
    more consistent with zipfile.ZipFile by raising a KeyError if an
    entry does not exist. It also provides a closing context for use in
    ``with`` statements.
    """

    def open(self, name, mode='rb'):
        if not os.path.isfile(os.path.join(self.fn, name)):
            raise KeyError("There is no item named '{}' in the archive".format(name))
        return super(ExplodedZipFile, self).open(name, mode)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, traceback):
        self.close()
        return exc_type is None



import os
import subprocess
import sys

from sphinx.errors import ExtensionError

from . import ref

def setup(app):
    """Register the extension with Sphinx.

    Args:
        app: The Sphinx application.
    """

    for name, (default, rebuild, _) in ref.CONFIG_VALUES.iteritems():
        app.add_config_value(name, default, rebuild)

    app.add_directive('javaimport', ref.JavarefImportDirective)
    app.add_role('javaref', ref.JavarefRole(app))

    # initialize_package_list must happen after validate_env
    app.connect('builder-inited', validate_env)
    app.connect('builder-inited', ref.initialize_package_list)

    app.connect('env-purge-doc', ref.purge_imports)


def validate_env(app):
    """Purge expired values from the environment.

    When certain configuration values change, related values in the
    environment must be cleared. While Sphinx can rebuild documents on
    configuration changes, it does not notify extensions when this
    happens. Instead, cache relevant values in the environment in order
    to detect when they change.

    Args:
        app: The Sphinx application.
    """

    if not hasattr(app.env, 'javalink_config_cache'):
        app.env.javalink_config_cache = {}

    for conf_attr, (_, _, env_attr) in ref.CONFIG_VALUES.iteritems():
        if not env_attr:
            continue

        value = getattr(app.config, conf_attr)
        cached = app.env.javalink_config_cache.get(conf_attr, value)

        app.env.javalink_config_cache[conf_attr] = value
        if value != cached:
            app.verbose('[javalink] config.%s has changed, clearing related env', conf_attr)
            delattr(app.env, env_attr)


def find_rt_jar(javahome=None):
    """Find the path to the Java standard library jar.

    The jar is expected to exist at the path 'jre/lib/rt.jar' inside a
    standard Java installation directory. The directory is found using
    the following procedure:

    1. If the javehome argument is provided, use the value as the
       directory.
    2. If the JAVA_HOME environment variable is set, use the value as
       the directory.
    3. Find the location of the ``java`` binary in the current PATH and
       compute the installation directory from this location.

    Args:
        javahome: A path to a Java installation directory (optional).
    """

    if not javahome:
        if 'JAVA_HOME' in os.environ:
            javahome = os.environ['JAVA_HOME']
        elif sys.platform == 'darwin':
            # The default java binary on OS X is not part of a standard Oracle
            # install, so building paths relative to it does not work like it
            # does on other platforms.
            javahome = _find_osx_javahome()
        else:
            javahome = _get_javahome_from_java(_find_java_binary())

    rtpath = os.path.join(javahome, 'jre', 'lib', 'rt.jar')
    if not os.path.isfile(rtpath):
        msg = 'Could not find rt.jar: {} is not a file'.format(rtpath)
        raise ExtensionError(msg)

    return rtpath


def _find_osx_javahome():
    return subprocess.check_output(['/usr/libexec/java_home']).strip()


def _get_javahome_from_java(java):
    while os.path.islink(java):
        link = java
        java = os.readlink(link)
        if not os.path.isabs(java):
            java = os.path.join(os.path.dirname(link), java)

    javahome = os.path.join(os.path.dirname(java), '..', '..')
    return os.path.normpath(javahome)


def _find_java_binary():
    path = os.environ.get('PATH', os.defpath)

    exts = os.environ.get('PATHEXT', '')
    java_names = ['java' + ext for ext in exts.split(os.pathsep)]

    for p in path.split(os.pathsep):
        p = p.strip('"')
        for name in java_names:
            java = os.path.join(p, name)
            if os.path.isfile(java) and os.access(java, os.X_OK):
                return java

    raise ExtensionError("Could not find 'java' binary in PATH")



import contextlib
import os
import traceback
import urllib2

import docutils.nodes
import docutils.utils

from urllib import quote as urlquote, pathname2url
from urlparse import urlparse, urlunparse, urljoin

from docutils.parsers import rst
from sphinx.util.nodes import split_explicit_title

from .loader import ClassLoader
from .model import parse_name


CONFIG_VALUES = {
    'javalink_classpath': ([], 'env', 'javalink_classloader'),
    'javalink_docroots': ([], 'env', 'javalink_packages'),
    'javalink_add_package_names': (True, 'env', None),
    'javalink_qualify_nested_types': (True, 'env', None),
    'javalink_add_method_parameters': (True, 'env', None)
}


def abspath(root, path):
    return os.path.normpath(os.path.join(root, path))


class EnvAccessor(object):
    @property
    def env(self):
        pass

    @property
    def classloader(self):
        if not hasattr(self.env, 'javalink_classloader'):
            classpath = self.env.config.javalink_classpath
            loader = ClassLoader([abspath(self.env.srcdir, p) for p in classpath])
            self.env.javalink_classloader = loader

        return self.env.javalink_classloader

    @property
    def imports(self):
        if not hasattr(self.env, 'javalink_imports'):
            self.env.javalink_imports = {}

        return self.env.javalink_imports


class JavarefImportDirective(rst.Directive, EnvAccessor):
    required_arguments = 0
    optional_arguments = 0
    has_content = True

    @property
    def env(self):
        return self.state.document.settings.env

    def run(self):
        docname = self.env.docname
        imports = self.imports.setdefault(docname, [('java.lang', '*')])

        for i in self.content:
            package, name = parse_name(i.strip())
            if (package.name, name) not in imports:
                # TODO make this work with Package objects
                self._validate_import(package.name, name)
                imports.append((package.name, name))

        return []

    # TODO handle conflicting imports
    def _validate_import(self, package, name):
        if name == '*':
            entity = self.classloader.find_package(package)
        else:
            entity = self.classloader.load('{}.{}'.format(package, name))

        if not entity:
            self.error("unresolved import '{}.{}'".format(package, name))


def purge_imports(app, env, docname):
    if hasattr(env, 'javalink_imports'):
        env.javalink_imports.pop(docname, None)


class JavarefRole(EnvAccessor):
    def __init__(self, app):
        self.app = app

    @property
    def env(self):
        return self.app.env

    def find_ref(self, reftext):
        reftext = reftext.strip()

        # TODO add additional validation (see SeeTagImpl.java)
        where, _, what = reftext.partition('#')
        clazz = self._find_class(where)
        if clazz:
            where = clazz.full_name
            if what:
                # TODO handle member ambiguity
                member = clazz.get_member(what)
                if not member:
                    raise JavarefError('unknown member: {}'.format(reftext))

                what = member.get_url_fragment()

            return where, what

        if not what:
            package = self.classloader.find_package(where)
            if package:
                return package.name + '.package-summary', None

        raise JavarefError('reference not found: {}'.format(reftext))

    def to_url(self, where, what):
        root = self._find_url_root(where)
        if not root:
            raise JavarefError('root URL not found: {}'.format(where))

        path = where.replace('.', '/').replace('$', '.')
        path += '.html'
        if what:
            # TODO is this the correct way to escape the URL?
            path += '#{}'.format(urlquote(what, ';/?:@&=+$,()'))

        return urljoin(root, path)

    def to_title(self, where, what):
        package, name = parse_name(where)
        if name == 'package-summary':
            return package.name

        add_package_names = self.app.config.javalink_add_package_names
        qualify_nested_types = self.app.config.javalink_qualify_nested_types
        add_method_parameters = self.app.config.javalink_add_method_parameters

        title = []
        if add_package_names:
            title.append(package.name)

        if add_package_names or qualify_nested_types:
            title.append(name.replace('$', '.'))
        else:
            title.append(name.rpartition('$')[-1])

        if what:
            if add_method_parameters:
                title.append(what)
            else:
                title.append(what.partition('(')[0])

        return '.'.join(title)

    def _find_class(self, where):
        import_name = where.partition('.')[0]
        imports = self.imports.get(self.env.docname, [])

        candidates = [where]
        for package, name in imports:
            if name == import_name or name == '*':
                candidates.append('{}.{}'.format(package, where))

        for name in candidates:
            clazz = self.classloader.load(name)
            if clazz:
                return clazz

            # try all possible class nestings
            sep = name.rfind('.')
            while sep >= 0:
                name = name[:sep] + '$' + name[sep+1:]
                clazz = self.classloader.load(name)
                if clazz:
                    return clazz

                sep = name.rfind('.', 0, sep)

        return None

    def _find_url_root(self, where):
        package, _ = parse_name(where)
        return self.env.javalink_packages.get(package.name, None)

    def __call__(self, name, rawtext, text, lineno, inliner,
                 options={}, content=[]):

        text = docutils.utils.unescape(text)
        has_title, title, reftext = split_explicit_title(text)

        warnings = []
        try:
            where, what = self.find_ref(reftext)
            url = self.to_url(where, what)
            if not has_title:
                title = self.to_title(where, what)
        except JavarefError as e:
            url = None
            warnings.append(e.reason)

        if url:
            # if no scheme, assume a local path relative to the src root
            if not urlparse(url).scheme:
                docdir = os.path.dirname(inliner.document.current_source)
                if docdir != self.env.srcdir:
                    url = os.path.relpath(self.env.srcdir, docdir) + '/' + url

            ref = docutils.nodes.reference('', '', internal=False, refuri=url)
            ref.append(docutils.nodes.literal(rawsource=title, text=title))
        else:
            ref = docutils.nodes.literal(rawsource=title, text=title)

        return [ref], [inliner.reporter.warning(w, line=lineno) for w in warnings]


class JavarefError(Exception):
    """Raised when a reference to a Java element cannot be resolved.

    Attributes:
        reason: reason why the reference is unresolved
    """

    def __init__(self, reason):
        self.reason = reason

    def __str__(self):
        return str(self.reason)


def initialize_package_list(app):
    env = app.env
    if hasattr(env, 'javalink_packages'):
        return

    app.verbose('[javalink] initializing package list...')
    env.javalink_packages = {}

    for url, base in [parse_docroot(env.srcdir, r) for r in app.config.javalink_docroots]:
        try:
            with contextlib.closing(urllib2.urlopen(url)) as package_list:
                for package in package_list:
                    package = package.strip()
                    if package not in env.javalink_packages:
                        env.javalink_packages[package] = base
                    else:
                        app.warn("[javalink] duplicate package '{}' in {}".format(package, url))

        except urllib2.URLError:
            app.warn('[javalink] could not get {}; some links may not resolve'.format(url))
            app.verbose('[javalink] %s', traceback.format_exc())


def parse_docroot(srcdir, root):
    """Creates a package-list URL and a link base from a docroot element.

    Args:
        srcdir: the Sphinx source directory
        root: the docroot element [string or tuple]
    """

    if isinstance(root, basestring):
        return _parse_docroot_str(srcdir, root)
    else:
        url = _parse_docroot_str(srcdir, root[0])[0]
        base = _parse_docroot_str(srcdir, root[1])[1]
        return (url, base)


def _parse_docroot_str(srcdir, root):
    scheme, netloc, path = urlparse(root)[0:3]
    if not scheme:
        # assume local path; add trailing '/'s if missing
        root = os.path.join(root, '')
        absroot = os.path.join(abspath(srcdir, root), '')

        url = urljoin('file:///', pathname2url(absroot))
        base = pathname2url(root)
    else:
        path = path.rstrip('/') + '/'

        url = urlunparse((scheme, netloc, path, '', '', ''))
        base = url

    return (urljoin(url, 'package-list'), base)
